import abc
from typing import Optional


class Aggregation(abc.ABC):
    def __init__(self, field: str) -> None:
        self.field = field

    @abc.abstractmethod
    def compile(self, depth: int = 1):
        ...


class MetricAggregation(Aggregation):
    ...


class BucketAggregation(Aggregation):
    @abc.abstractmethod
    def nested(self, child: Aggregation):
        ...


class Terms(BucketAggregation):
    def __init__(self, field: str, max_buckets: int = 100) -> None:
        super().__init__(field)
        self.max_buckets = max_buckets
        self.child: Optional[Aggregation] = None

    def compile(self, depth: int = 1):
        return {
            depth: {
                'terms': {
                    'field': self.field,
                    'size': self.max_buckets,
                },
                'aggs': self.child.compile(depth + 1) if self.child else {},
            }
        }

    def nested(self, child: Aggregation):
        self.child = child
        return self


class Cardinality(MetricAggregation):
    def compile(self, depth: int = 1):
        return {
            depth: {
                'cardinality': {
                    'field': self.field,
                }
            }
        }


class Sum(MetricAggregation):
    def compile(self, depth: int = 1):
        return {
            depth: {
                'sum': {
                    'field': self.field,
                }
            }
        }



def parse_aggregations(data: dict, depth: int = 1):
    level = data.get(str(depth), None)
    if level is None:
        return

    if 'buckets' in level:
        result = {}
        buckets = level['buckets']
        for b in buckets:
            key = b['key']
            count = b['doc_count']
            children = parse_aggregations(b, depth + 1)
            result[key] = children if children else count
        return result
    else:
        value = level['value']
        return value



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['opensearchorm']

package_data = \
{'': ['*']}

install_requires = \
['opensearch-py>=2.0.0,<3.0.0',
 'pydantic>=1.10.2,<2.0.0',
 'pytz>=2022.2.1,<2023.0.0']

setup_kwargs = {
    'name': 'opensearch-orm',
    'version': '0.1.4',
    'description': '',
    'long_description': "# OpenSearch ORM\n`opensearch-orm` is a high-level OpenSearch ORM for Python. The query syntax is similar to django-orm.\n\nMay be compatible with Elasticsearch, depending on opensearch-py.\n\n\n# Installation\n``` bash\npip install opensearch-orm\n```\n\n\n# Getting Started\n\nFirst, define your document model with indexing pattern.\n``` python\nfrom opensearchorm import SearchSession, BaseModel\n\n\nclass UserLog(BaseModel):\n    __index__ = 'user_access_log-*'\n\n    method: str\n    path: str\n    remote_ip: str\n    created: datetime\n```\n\n\nYou can use django-like syntax or typed query expressions together.\n## filter\n``` python\n# {'bool': {'must_not': [], 'should': [], 'filter': [{'range': {'created': {'gte': '2022-09-01'}}}, {'match_phrase': {'remote_ip': '127.0.0.1'}}]}}        \nwith SearchSession() as session:\n    result = (\n        session.select(UserLog)\n        .filter(created__gte='2022-09-01', remote_ip='127.0.0.1')\n        .fetch()\n    )\n    print(result)\n\n    # equals to\n    result = (\n        session.select(UserLog)\n        .filter(Range('created', date(2022, 9, 1)), remote_ip='127.0.0.1')\n        .fetch()\n    )\n```\n## contains\n``` python\n# {'bool': {'must_not': [], 'should': [], 'filter': [{'bool': {'should': [{'match_phrase': {'method': 'GET'}}, {'match_phrase': {'method': 'POST'}}], 'minimum_should_match': 1}}]}}      \nwith SearchSession() as session:\n    result = (\n        session.select(UserLog)\n        .filter(method__contains=['GET', 'POST'])\n        .fetch()\n    )\n    print(result)\n\n    # equals to\n    result = (\n        session.select(UserLog)\n        .filter(Contains('method', ['GET', 'POST']))\n        .fetch()\n    )\n```\n\n## exclude\n``` python\n# {'bool': {'must_not': [{'match_phrase': {'method': 'get'}}, {'match_phrase': {'path': '/login'}}], 'should': [], 'filter': []}}\nwith SearchSession() as session:\n    result = (\n        session.select(UserLog)\n        .exclude(method='get', path='/login')\n        .fetch()\n    )\n    print(result)\n```\n\n\n## paginate\n``` python\nwith SearchSession() as session:\n    result = (\n        session.select(UserLog)\n        .filter(method='get')\n        .limit(100)\n        .offset(100)\n        .fetch()\n    )\n    print(result)\n```\n\n## aggregations\ngroup by path and count unique remote_ip.\n\n``` python\nwith SearchSession() as session:\n    # aggregate text must need use a keyword field instead\n    # request_timeout argument will be passed on to the opensearch-py\n    result = (\n        session.select(UserLog)\n        .aggregate(Terms('path.keyword').nested(Cardinality('remote_ip,keyword')), request_timeout=300)\n    )\n    print(result)\n    # result -> {'path': 1, 'path2': 2}\n```",
    'author': 'yim7',
    'author_email': 'yimchiu7@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import ClassVar, Optional
from pydantic import BaseModel as RawBaseModel


class BaseModel(RawBaseModel):
    __index__: ClassVar[Optional[str]] = None

    @classmethod
    def default_fields(cls):
        return list(cls.__fields__.keys())



import abc
from datetime import date, datetime
from enum import Enum
import logging
from typing import Any, Callable, Dict, List, Tuple, Type, TypeVar, Union

from opensearchorm.model import BaseModel

Model = TypeVar('Model', bound=BaseModel)


class Expr(abc.ABC):
    @abc.abstractmethod
    def compile(self) -> dict:
        ...


class Contains(Expr):
    def __init__(self, field: str, values: list, min_match: int = 1):
        self.field = field
        self.values = values
        self.min_match = min_match

    def compile(self):
        return {
            'bool': {
                'should': [MatchPhrase(self.field, v).compile() for v in self.values],
                'minimum_should_match': self.min_match,
            }
        }


class Range(Expr):
    def __init__(self, field: str, interval: Tuple[Any, Any], *, left_open: bool = False, right_open: bool = False):
        self.field = field
        self.interval = interval
        self.left_open = left_open
        self.right_open = right_open

    def compile(self):
        left, right = self.interval
        if isinstance(left, (date, datetime)):
            left = left.isoformat()
        if isinstance(right, (date, datetime)):
            right = right.isoformat()

        range = {}
        if left is not None:
            op = 'gt' if self.left_open else 'gte'
            range[op] = left
        if right is not None:
            op = 'lt' if self.right_open else 'lte'
            range[op] = right

        return {
            'range': {
                self.field: range,
            }
        }


class MatchPhrase(Expr):
    def __init__(self, field: str, value: str):
        self.field = field
        self.value = value

    def compile(self):
        return {
            'match_phrase': {
                self.field: self.value,
            }
        }


class MatchPhrasePrefix(Expr):
    def __init__(self, field: str, value: str):
        self.field = field
        self.value = value

    def compile(self):
        return {
            'match_phrase_prefix': {
                self.field: self.value,
            }
        }


class Wildcard(Expr):
    def __init__(self, field: str, value: str):
        self.field = field
        self.value = value

    def compile(self):
        return {
            'wildcard': {
                self.field: self.value,
            }
        }


class RegExp(Expr):
    def __init__(self, field: str, value: str):
        self.field = field
        self.value = value

    def compile(self):
        return {
            'regexp': {
                self.field: self.value,
            }
        }


class Operator(Enum):
    PREFIX = '__prefix'
    REGEXP = '__regexp'
    CONTAINS = '__contains'
    GTE = '__gte'
    GT = '__gt'
    LTE = '__lte'
    LT = '__lt'


OPERATOR_FUNCTIONS: Dict[Operator, Callable[[str, Any], Expr]] = {
    Operator.CONTAINS: lambda field, value: Contains(field, value),
    Operator.PREFIX: lambda field, value: MatchPhrasePrefix(field, value),
    Operator.REGEXP: lambda field, value: RegExp(field, value),
    Operator.GTE: lambda field, value: Range(field, (value, None)),
    Operator.GT: lambda field, value: Range(field, (value, None), left_open=True),
    Operator.LTE: lambda field, value: Range(field, (None, value)),
    Operator.LT: lambda field, value: Range(field, (None, value), right_open=True),
}


class ModelQuery(Expr):
    def __init__(self, model_cls: Type[Model]):
        self.__model_cls = model_cls
        self.__filter: List[Expr] = []
        self.__exclude: List[Expr] = []
        self.__union: List[Expr] = []

    def compile(self):
        return {
            'bool': {
                'must_not': [e.compile() for e in self.__exclude],
                'should': [e.compile() for e in self.__union],
                'filter': [e.compile() for e in self.__filter],
                'minimum_should_match': 1 if self.__union else 0,
            }
        }

    @property
    def valid_fields(self):
        # todo @functools.cached_property, not supported in python3.7
        return set(self.__model_cls.default_fields())

    def check_valid_field(self, field: str):
        assert field in self.valid_fields, f'check field name: {field}'

    def parse_clause(self, raw_field: str, value) -> Expr:
        field = raw_field
        for op in Operator:
            suffix: str = op.value
            if raw_field.endswith(suffix):
                field, _ = raw_field.rsplit(suffix)
                logging.debug('parse field: %s, raw: %s', field, raw_field)
                self.check_valid_field(field)
                return OPERATOR_FUNCTIONS[op](field, value)

        self.check_valid_field(field)
        return MatchPhrase(field, value)

    def parse_clauses(self, **kwargs):
        clauses = []
        for k, v in kwargs.items():
            cond = self.parse_clause(k, v)
            clauses.append(cond)
        return clauses

    def filter(self, *args: Expr, **kwargs):
        conditions = self.parse_clauses(**kwargs)
        self.__filter.extend(args)
        self.__filter.extend(conditions)

        return self

    def union(self, *args: Expr, **kwargs):
        conditions = self.parse_clauses(**kwargs)
        self.__union.extend(args)
        self.__union.extend(conditions)

        return self

    def exclude(self, *args: Expr, **kwargs):
        conditions = self.parse_clauses(**kwargs)
        self.__exclude.extend(args)
        self.__exclude.extend(conditions)

        return self
